# standard libraries
import contextlib
import logging
import math
import unittest

# third party libraries
import numpy

# local libraries
from nion.data import Calibration
from nion.data import DataAndMetadata
from nion.swift import Application
from nion.swift import DocumentController
from nion.swift import LineGraphCanvasItem
from nion.swift.model import DataItem
from nion.swift.model import DocumentModel
from nion.swift.model import Graphics
from nion.ui import TestUI


class TestLineGraphCanvasItem(unittest.TestCase):

    def setUp(self):
        self.app = Application.Application(TestUI.UserInterface(), set_global=False)

    def tearDown(self):
        pass

    def test_data_values_give_pretty_limits_when_auto(self):

        test_ranges = (
            ((7.46, 85.36), (0.0, 100.0)),
            ((7.67, 12.95), (0.0, 15.0)),
            ((6.67, 11.95), (0.0, 15.0)),
            ((0.00, 0.00), (0.0, 0.0))
        )

        for data_in, data_out in test_ranges:
            data_min, data_max = data_in
            expected_uncalibrated_data_min, expected_uncalibrated_data_max = data_out
            data = numpy.zeros((16, 16), dtype=numpy.float64)
            irow, icol = numpy.ogrid[0:16, 0:16]
            data[:] = data_min + (data_max - data_min) * (irow / 15.0)
            # auto on min/max
            calibrated_data_min, calibrated_data_max, y_ticker = LineGraphCanvasItem.calculate_y_axis([data], None, None, None, None)
            axes = LineGraphCanvasItem.LineGraphAxes(1.0, calibrated_data_min, calibrated_data_max, y_ticker=y_ticker)
            self.assertEqual(axes.uncalibrated_data_min, expected_uncalibrated_data_min)
            self.assertEqual(axes.uncalibrated_data_max, expected_uncalibrated_data_max)

    def test_display_limits_are_reasonable_when_using_log_scale(self):
        data = numpy.linspace(-0.1, 10.0, 10)
        data_style = "log"
        calibrated_data_min, calibrated_data_max, y_ticker = LineGraphCanvasItem.calculate_y_axis([data], None, None, None, data_style)
        axes = LineGraphCanvasItem.LineGraphAxes(1.0, calibrated_data_min, calibrated_data_max, data_style=data_style, y_ticker=y_ticker)
        self.assertAlmostEqual(axes.uncalibrated_data_min, 1.0)
        self.assertAlmostEqual(axes.uncalibrated_data_max, 10.0)
        calibrated_data = axes.calculate_calibrated_xdata(DataAndMetadata.new_data_and_metadata(data)).data
        self.assertAlmostEqual(numpy.amin(calibrated_data), math.log10(1.0))
        self.assertAlmostEqual(numpy.amax(calibrated_data), math.log10(10.0))

    def test_display_limits_are_reasonable_when_using_calibrated_log_scale(self):
        intensity_calibration = Calibration.Calibration(-5, 2)
        data = numpy.linspace(-0.1, 10.0, 10)
        data_style = "log"
        calibrated_data_min, calibrated_data_max, y_ticker = LineGraphCanvasItem.calculate_y_axis([data], None, None, intensity_calibration, data_style)
        axes = LineGraphCanvasItem.LineGraphAxes(1.0, calibrated_data_min, calibrated_data_max, y_calibration=intensity_calibration, data_style=data_style, y_ticker=y_ticker)
        self.assertAlmostEqual(axes.calibrated_data_min, 0.0)
        self.assertAlmostEqual(axes.calibrated_data_max, 1.5)  # empirically mesaured
        calibrated_data = axes.calculate_calibrated_xdata(DataAndMetadata.new_data_and_metadata(data)).data
        self.assertAlmostEqual(numpy.amin(calibrated_data), math.log10(1.0))
        self.assertAlmostEqual(numpy.amax(calibrated_data), math.log10(15.0))

    def test_tool_returns_to_pointer_after_but_not_during_creating_interval(self):
        # setup
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((100,)))
            document_model.append_data_item(data_item)
            display_panel.set_display_panel_data_item(data_item)
            display_panel.display_canvas_item.layout_immediate((800, 1000))
            # run test
            self.assertEqual(document_controller.tool_mode, "pointer")
            for tool_mode in ["interval"]:
                document_controller.tool_mode = tool_mode
                display_panel.display_canvas_item.simulate_press((100,125))
                display_panel.display_canvas_item.simulate_move((100,125))
                self.assertEqual(document_controller.tool_mode, tool_mode)
                display_panel.display_canvas_item.simulate_move((250,200))
                display_panel.display_canvas_item.simulate_release((250,200))
                self.assertEqual(document_controller.tool_mode, "pointer")

    def test_pointer_tool_makes_intervals(self):
        # setup
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((100,)))
            document_model.append_data_item(data_item)
            display_panel.set_display_panel_data_item(data_item)
            display_panel.display_canvas_item.layout_immediate((640, 480))
            # test
            document_controller.tool_mode = "pointer"
            display_panel.display_canvas_item.simulate_drag((240, 160), (240, 480))
            interval_region = data_item.displays[0].graphics[0]
            self.assertEqual(interval_region.type, "interval-graphic")
            self.assertTrue(interval_region.end > interval_region.start)

    def test_pointer_tool_makes_intervals_when_other_intervals_exist(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((100,)))
            region = Graphics.IntervalGraphic()
            region.start = 0.9
            region.end = 0.95
            data_item.displays[0].add_graphic(region)
            document_model.append_data_item(data_item)
            display_panel.set_display_panel_data_item(data_item)
            display_panel.display_canvas_item.layout_immediate((640, 480))
            # test
            document_controller.tool_mode = "pointer"
            display_panel.display_canvas_item.simulate_drag((240, 160), (240, 480))
            interval_region = data_item.displays[0].graphics[1]
            self.assertEqual(interval_region.type, "interval-graphic")
            self.assertTrue(interval_region.end > interval_region.start)

    def test_nudge_interval(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((100,)))
            region = Graphics.IntervalGraphic()
            region.start = 0.1
            region.end = 0.9
            data_item.displays[0].add_graphic(region)
            document_model.append_data_item(data_item)
            display_panel.set_display_panel_data_item(data_item)
            display_panel.display_canvas_item.layout_immediate((640, 480))
            display_panel.display_canvas_item.prepare_display()  # force layout
            # test
            document_controller.tool_mode = "pointer"
            display_panel.display_canvas_item.simulate_click((240, 320))
            display_panel.display_canvas_item.key_pressed(self.app.ui.create_key_by_id("left"))
            interval_region = data_item.displays[0].graphics[0]
            self.assertTrue(interval_region.start < 0.1)
            self.assertTrue(interval_region.end < 0.9)
            self.assertAlmostEqual(interval_region.end - interval_region.start, 0.8)

    def test_line_plot_auto_scales_uncalibrated_y_axis(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            display_panel = document_controller.selected_display_panel
            data = numpy.zeros((100,))
            data[50] = 75
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            display_panel.set_display_panel_data_item(data_item)
            display_panel.display_canvas_item.layout_immediate((640, 480))
            axes = display_panel.display_canvas_item._axes
            self.assertAlmostEqual(axes.calibrated_data_min, 0.0)
            self.assertAlmostEqual(axes.calibrated_data_max, 80.0)
            self.assertAlmostEqual(axes.uncalibrated_data_min, 0.0)
            self.assertAlmostEqual(axes.uncalibrated_data_max, 80.0)

    def test_line_plot_auto_scales_calibrated_y_axis(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            display_panel = document_controller.selected_display_panel
            data = numpy.zeros((10,))
            data[5] = 75
            data_item = DataItem.DataItem(data)
            data_item.set_xdata(DataAndMetadata.new_data_and_metadata(data, Calibration.Calibration(0, 0.5, "x")))
            document_model.append_data_item(data_item)
            display_panel.set_display_panel_data_item(data_item)
            display_panel.display_canvas_item.layout_immediate((640, 480))
            axes = display_panel.display_canvas_item._axes
            self.assertAlmostEqual(axes.calibrated_data_min, 0.0)
            self.assertAlmostEqual(axes.calibrated_data_max, 40.0)
            self.assertAlmostEqual(axes.uncalibrated_data_min, 0.0)
            self.assertAlmostEqual(axes.uncalibrated_data_max, 80.0)

    def test_line_plot_with_no_data_displays_gracefully(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.ones((8,), numpy.float))
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            display_specifier.data_item.set_xdata(DataAndMetadata.DataAndMetadata(lambda: None, ((8, 0), numpy.float)))
            display_specifier.display.display_type = "line_plot"
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(data_item)
            display_panel.display_canvas_item.layout_immediate((640, 480))

    def test_composite_line_plot_initializes_properly(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item1 = DataItem.DataItem(numpy.ones((8,), numpy.float))
            data_item2 = DataItem.DataItem(numpy.ones((8,), numpy.float))
            document_model.append_data_item(data_item1)
            document_model.append_data_item(data_item2)
            composite_item = DataItem.CompositeLibraryItem()
            composite_item.append_data_item(data_item1)
            composite_item.append_data_item(data_item2)
            composite_item.displays[0].display_type = "line_plot"
            document_model.append_data_item(composite_item)
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(composite_item)
            display_panel.display_canvas_item.layout_immediate((640, 480))

    def test_composite_line_plot_calculates_calibrated_data_of_two_data_items_with_same_units_but_different_scales_properly(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            xdata1 = DataAndMetadata.new_data_and_metadata(numpy.ones((8,)), dimensional_calibrations=[Calibration.Calibration(offset=0, scale=1, units="nm")])
            xdata2 = DataAndMetadata.new_data_and_metadata(numpy.ones((4,)), dimensional_calibrations=[Calibration.Calibration(offset=2, scale=2, units="nm")])
            data_item1 = DataItem.new_data_item(xdata1)
            data_item2 = DataItem.new_data_item(xdata2)
            document_model.append_data_item(data_item1)
            document_model.append_data_item(data_item2)
            composite_item = DataItem.CompositeLibraryItem()
            composite_item.append_data_item(data_item1)
            composite_item.append_data_item(data_item2)
            composite_item.displays[0].display_type = "line_plot"
            document_model.append_data_item(composite_item)
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(composite_item)
            display_panel.display_canvas_item.layout_immediate((640, 480))
            # print(display_panel.display_canvas_item.line_graph_stack.canvas_items[0].calibrated_data)
            # print(display_panel.display_canvas_item.line_graph_stack.canvas_items[1].calibrated_data)

    def test_composite_line_plot_handles_drawing_with_fixed_y_scale_and_without_data(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            composite_item = DataItem.CompositeLibraryItem()
            composite_item.displays[0].display_type = "line_plot"
            composite_item.displays[0].y_min = 0
            composite_item.displays[0].y_max = 1
            document_model.append_data_item(composite_item)
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(composite_item)
            display_panel.display_canvas_item.layout_immediate((640, 480))
            display_panel.display_canvas_item.prepare_display()  # force layout

    def test_composite_line_plot_handles_first_components_without_data(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item1 = DataItem.DataItem()
            data_item2 = DataItem.DataItem(numpy.ones((8,), numpy.float))
            document_model.append_data_item(data_item1)
            document_model.append_data_item(data_item2)
            composite_item = DataItem.CompositeLibraryItem()
            composite_item.append_data_item(data_item1)
            composite_item.append_data_item(data_item2)
            composite_item.displays[0].display_type = "line_plot"
            document_model.append_data_item(composite_item)
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(composite_item)
            display_panel.display_canvas_item.layout_immediate((640, 480))
            display_panel.display_canvas_item.prepare_display()  # force layout

    def test_line_plot_calculates_calibrated_vs_uncalibrated_display_y_values(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.new_data_item(DataAndMetadata.new_data_and_metadata(numpy.ones((8, )), intensity_calibration=Calibration.Calibration(offset=0, scale=10, units="nm")))
            data_item.displays[0].display_type = "line_plot"
            document_model.append_data_item(data_item)
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(data_item)
            display_panel.display_canvas_item.layout_immediate((640, 480))
            self.assertTrue(numpy.array_equal(display_panel.display_canvas_item.line_graph_canvas_item.calibrated_xdata.data, numpy.full((8, ), 10)))
            data_item.displays[0].dimensional_calibration_style = "pixels-top-left"
            display_panel.display_canvas_item.layout_immediate((640, 480))
            self.assertTrue(numpy.array_equal(display_panel.display_canvas_item.line_graph_canvas_item.calibrated_xdata.data, numpy.ones((8, ))))

    def test_line_plot_handles_calibrated_vs_uncalibrated_display(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.new_data_item(DataAndMetadata.new_data_and_metadata(numpy.ones((8, )), dimensional_calibrations=[Calibration.Calibration(offset=0, scale=10, units="nm")]))
            data_item.displays[0].display_type = "line_plot"
            document_model.append_data_item(data_item)
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(data_item)
            display_panel.display_canvas_item.layout_immediate((640, 480))
            self.assertEqual(display_panel.display_canvas_item.line_graph_canvas_item.calibrated_xdata.dimensional_calibrations[-1].units, "nm")
            data_item.displays[0].dimensional_calibration_style = "pixels-top-left"
            display_panel.display_canvas_item.layout_immediate((640, 480))
            self.assertFalse(display_panel.display_canvas_item.line_graph_canvas_item.calibrated_xdata.dimensional_calibrations[-1].units)

    def test_multi_line_plot_without_calibration_does_not_display_any_line_graphs(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.new_data_item(DataAndMetadata.new_data_and_metadata(numpy.ones((8, )), dimensional_calibrations=[Calibration.Calibration(offset=0, scale=10, units="nm")]))
            data_item.displays[0].display_type = "line_plot"
            document_model.append_data_item(data_item)
            composite_item = DataItem.CompositeLibraryItem()
            composite_item.append_data_item(data_item)
            composite_item.displays[0].display_type = "line_plot"
            document_model.append_data_item(composite_item)
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(composite_item)
            display_panel.display_canvas_item.layout_immediate((640, 480))
            self.assertIsNone(display_panel.display_canvas_item.line_graph_canvas_item)

    def test_multi_line_plot_handles_calibrated_vs_uncalibrated_display(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            calibration = Calibration.Calibration(offset=0, scale=10, units="nm")
            data_item = DataItem.new_data_item(DataAndMetadata.new_data_and_metadata(numpy.ones((8, )), dimensional_calibrations=[calibration]))
            data_item.displays[0].display_type = "line_plot"
            document_model.append_data_item(data_item)
            composite_item = DataItem.CompositeLibraryItem()
            composite_item.append_data_item(data_item)
            composite_item.displays[0].display_type = "line_plot"
            composite_item.displays[0].dimensional_calibrations = [calibration]
            composite_item.displays[0].dimensional_scales = [8]
            document_model.append_data_item(composite_item)
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(composite_item)
            display_panel.display_canvas_item.layout_immediate((640, 480))
            self.assertEqual(display_panel.display_canvas_item.line_graph_canvas_item.calibrated_xdata.dimensional_calibrations[-1].units, composite_item.displays[0].dimensional_calibrations[-1].units)
            # print(f"style {composite_item.displays[0].dimensional_calibration_style}")
            # print(f"dim {composite_item.displays[0].dimensional_calibrations}")
            # print(f"int {composite_item.displays[0].intensity_calibration}")
            # print(f"d dim {composite_item.displays[0].displayed_dimensional_calibrations}")
            # print(f"d int {composite_item.displays[0].displayed_intensity_calibration}")
            # print(f"scales {composite_item.displays[0].displayed_dimensional_scales}")
            composite_item.displays[0].dimensional_calibration_style = "pixels-top-left"
            display_panel.display_canvas_item.layout_immediate((640, 480))
            # print(f">> {display_panel.display_canvas_item.line_graph_canvas_item}")
            self.assertFalse(display_panel.display_canvas_item.line_graph_canvas_item.calibrated_xdata.dimensional_calibrations[-1].units)
            self.assertFalse(composite_item.displays[0].displayed_dimensional_calibrations[-1].units)
            # print(f"style {composite_item.displays[0].dimensional_calibration_style}")
            # print(f"dim {composite_item.displays[0].dimensional_calibrations}")
            # print(f"int {composite_item.displays[0].intensity_calibration}")
            # print(f"d dim {composite_item.displays[0].displayed_dimensional_calibrations}")
            # print(f"d int {composite_item.displays[0].displayed_intensity_calibration}")
            # print(f"scales {composite_item.displays[0].displayed_dimensional_scales}")


if __name__ == '__main__':
    logging.getLogger().setLevel(logging.DEBUG)
    unittest.main()



# standard libraries
import contextlib
import copy
import functools
import logging
import unittest

# third party libraries
import numpy

# local libraries
from nion.data import Calibration
from nion.data import DataAndMetadata
from nion.swift import Application
from nion.swift import DocumentController
from nion.swift import Facade
from nion.swift.model import Cache
from nion.swift.model import DataItem
from nion.swift.model import DocumentModel
from nion.swift.model import Graphics
from nion.ui import TestUI
from nion.utils import Geometry


Facade.initialize()


class TestProcessingClass(unittest.TestCase):

    def setUp(self):
        self.app = Application.Application(TestUI.UserInterface(), set_global=False)
        cache_name = ":memory:"
        storage_cache = Cache.DbStorageCache(cache_name)
        self.document_model = DocumentModel.DocumentModel(storage_cache=storage_cache)
        self.document_controller = DocumentController.DocumentController(self.app.ui, self.document_model, workspace_id="library")
        self.display_panel = self.document_controller.selected_display_panel
        self.data_item = DataItem.DataItem(numpy.zeros((10, 10)))
        self.display_specifier = DataItem.DisplaySpecifier.from_data_item(self.data_item)
        self.document_model.append_data_item(self.data_item)
        self.display_panel.set_display_panel_data_item(self.data_item)

    def tearDown(self):
        self.document_controller.close()

    # test processing against 1d data. doesn't test for correctness of the processing.
    def test_processing_1d(self):
        data_item_real = DataItem.DataItem(numpy.zeros((256), numpy.double))
        self.document_model.append_data_item(data_item_real)

        data_item_complex = DataItem.DataItem(numpy.zeros((256), numpy.complex128))
        self.document_model.append_data_item(data_item_complex)

        processing_list = []
        processing_list.append((data_item_real, self.document_model.get_fft_new, {}))
        processing_list.append((data_item_complex, self.document_model.get_ifft_new, {}))
        processing_list.append((data_item_real, self.document_model.get_invert_new, {}))
        processing_list.append((data_item_real, self.document_model.get_sobel_new, {}))
        processing_list.append((data_item_real, self.document_model.get_laplace_new, {}))
        processing_list.append((data_item_real, self.document_model.get_gaussian_blur_new, {}))
        processing_list.append((data_item_real, self.document_model.get_median_filter_new, {}))
        processing_list.append((data_item_real, self.document_model.get_uniform_filter_new, {}))
        processing_list.append((data_item_real, self.document_model.get_histogram_new, {}))
        processing_list.append((data_item_real, self.document_model.get_convert_to_scalar_new, {}))
        processing_list.append((data_item_real, self.document_model.get_transpose_flip_new, {"do_transpose": True, "do_flip_v": True, "do_flip_h": True}))

        for source_data_item, fn, params in processing_list:
            data_item = fn(source_data_item)
            for name, value in params.items():
                self.document_model.get_data_item_computation(data_item)._set_variable_value(name, value)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            self.document_model.recompute_all()
            with display_specifier.data_item.data_ref() as data_ref:
                src_data_item = self.document_model.resolve_object_specifier(self.document_model.get_data_item_computation(data_item).variables[0].variable_specifier).value.data_item
                self.assertEqual(src_data_item, source_data_item)
                self.assertIsNotNone(data_ref.data)
                self.assertIsNotNone(display_specifier.data_item.dimensional_calibrations)
                self.assertEqual(display_specifier.data_item.data_shape, data_ref.data.shape)
                self.assertEqual(display_specifier.data_item.data_dtype, data_ref.data.dtype)
                self.assertIsNotNone(display_specifier.data_item.data_dtype.type)  # make sure we're returning a dtype
                self.assertEqual(len(display_specifier.data_item.dimensional_shape), len(display_specifier.data_item.dimensional_calibrations))

    # test processing against 2d data. doesn't test for correctness of the processing.
    def test_processing_2d(self):
        data_item_real = DataItem.DataItem(numpy.zeros((8, 8), numpy.double))
        self.document_model.append_data_item(data_item_real)

        processing_list = []
        processing_list.append((data_item_real, self.document_model.get_fft_new, {}))
        processing_list.append((data_item_real, self.document_model.get_invert_new, {}))
        processing_list.append((data_item_real, self.document_model.get_sobel_new, {}))
        processing_list.append((data_item_real, self.document_model.get_laplace_new, {}))
        processing_list.append((data_item_real, self.document_model.get_auto_correlate_new, {}))
        processing_list.append((data_item_real, self.document_model.get_gaussian_blur_new, {}))
        processing_list.append((data_item_real, self.document_model.get_median_filter_new, {}))
        processing_list.append((data_item_real, self.document_model.get_uniform_filter_new, {}))
        processing_list.append((data_item_real, self.document_model.get_crop_new, {}))
        processing_list.append((data_item_real, self.document_model.get_transpose_flip_new, {"do_transpose": True, "do_flip_v": True, "do_flip_h": True}))
        processing_list.append((data_item_real, self.document_model.get_resample_new, {"width": 128, "height": 128}))
        processing_list.append((data_item_real, self.document_model.get_histogram_new, {}))
        processing_list.append((data_item_real, self.document_model.get_line_profile_new, {}))
        processing_list.append((data_item_real, self.document_model.get_projection_new, {}))
        processing_list.append((data_item_real, self.document_model.get_convert_to_scalar_new, {}))

        for source_data_item, fn, params in processing_list:
            data_item = fn(source_data_item)
            for name, value in params.items():
                self.document_model.get_data_item_computation(data_item)._set_variable_value(name, value)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            self.document_model.recompute_all()
            with display_specifier.data_item.data_ref() as data_ref:
                src_data_item = self.document_model.resolve_object_specifier(self.document_model.get_data_item_computation(data_item).variables[0].variable_specifier).value.data_item
                self.assertEqual(src_data_item, source_data_item)
                self.assertIsNotNone(data_ref.data)
                self.assertIsNotNone(display_specifier.data_item.dimensional_calibrations)
                self.assertEqual(display_specifier.data_item.data_shape, data_ref.data.shape)
                self.assertEqual(display_specifier.data_item.data_dtype, data_ref.data.dtype)
                self.assertIsNotNone(display_specifier.data_item.data_dtype.type)  # make sure we're returning a dtype
                self.assertEqual(len(display_specifier.data_item.dimensional_shape), len(display_specifier.data_item.dimensional_calibrations))

    # test processing against 2d data. doesn't test for correctness of the processing.
    def test_processing_3d(self):
        data_item_real = DataItem.DataItem(numpy.zeros((16,16,256), numpy.double))
        self.document_model.append_data_item(data_item_real)

        processing_list = []
        processing_list.append((data_item_real, self.document_model.get_slice_sum_new, {}))
        processing_list.append((data_item_real, self.document_model.get_pick_new, {}))
        processing_list.append((data_item_real, self.document_model.get_pick_region_new, {}))

        for source_data_item, fn, params in processing_list:
            data_item = fn(source_data_item)
            for name, value in params.items():
                self.document_model.get_data_item_computation(data_item)._set_variable_value(name, value)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            self.document_model.recompute_all()
            with display_specifier.data_item.data_ref() as data_ref:
                src_data_item = self.document_model.resolve_object_specifier(self.document_model.get_data_item_computation(data_item).variables[0].variable_specifier).value.data_item
                self.assertEqual(src_data_item, source_data_item)
                self.assertIsNotNone(data_ref.data)
                self.assertIsNotNone(display_specifier.data_item.dimensional_calibrations)
                self.assertEqual(display_specifier.data_item.data_shape, data_ref.data.shape)
                self.assertEqual(display_specifier.data_item.data_dtype, data_ref.data.dtype)
                self.assertIsNotNone(display_specifier.data_item.data_dtype.type)  # make sure we're returning a dtype
                self.assertEqual(len(display_specifier.data_item.dimensional_shape), len(display_specifier.data_item.dimensional_calibrations))

    # test processing against 2d data. doesn't test for correctness of the processing.
    def test_processing_2d_rgb(self):
        data_item_rgb = DataItem.DataItem(numpy.zeros((8, 8, 3), numpy.uint8))
        self.document_model.append_data_item(data_item_rgb)

        processing_list = []
        processing_list.append((data_item_rgb, self.document_model.get_invert_new, {}))
        processing_list.append((data_item_rgb, self.document_model.get_sobel_new, {}))
        processing_list.append((data_item_rgb, self.document_model.get_laplace_new, {}))
        processing_list.append((data_item_rgb, self.document_model.get_gaussian_blur_new, {}))
        processing_list.append((data_item_rgb, self.document_model.get_median_filter_new, {}))
        processing_list.append((data_item_rgb, self.document_model.get_uniform_filter_new, {}))
        processing_list.append((data_item_rgb, self.document_model.get_crop_new, {}))
        processing_list.append((data_item_rgb, self.document_model.get_transpose_flip_new, {"do_transpose": True, "do_flip_v": True, "do_flip_h": True}))
        processing_list.append((data_item_rgb, self.document_model.get_resample_new, {"width": 128, "height": 128}))
        processing_list.append((data_item_rgb, self.document_model.get_histogram_new, {}))
        processing_list.append((data_item_rgb, self.document_model.get_line_profile_new, {}))
        processing_list.append((data_item_rgb, self.document_model.get_projection_new, {}))
        processing_list.append((data_item_rgb, self.document_model.get_convert_to_scalar_new, {}))

        for source_data_item, fn, params in processing_list:
            data_item = fn(source_data_item)
            for name, value in params.items():
                self.document_model.get_data_item_computation(data_item)._set_variable_value(name, value)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            self.document_model.recompute_all()
            with display_specifier.data_item.data_ref() as data_ref:
                src_data_item = self.document_model.resolve_object_specifier(self.document_model.get_data_item_computation(data_item).variables[0].variable_specifier).value.data_item
                self.assertEqual(src_data_item, source_data_item)
                self.assertIsNotNone(data_ref.data)
                self.assertIsNotNone(display_specifier.data_item.dimensional_calibrations)
                self.assertEqual(display_specifier.data_item.data_shape, data_ref.data.shape)
                self.assertEqual(display_specifier.data_item.data_dtype, data_ref.data.dtype)
                self.assertIsNotNone(display_specifier.data_item.data_dtype.type)  # make sure we're returning a dtype
                self.assertEqual(len(display_specifier.data_item.dimensional_shape), len(display_specifier.data_item.dimensional_calibrations))

    # test processing against 2d data. doesn't test for correctness of the processing.
    def test_processing_2d_rgba(self):
        data_item_rgb = DataItem.DataItem(numpy.zeros((8, 8, 4), numpy.uint8))
        self.document_model.append_data_item(data_item_rgb)

        processing_list = []
        processing_list.append((data_item_rgb, self.document_model.get_invert_new, {}))
        processing_list.append((data_item_rgb, self.document_model.get_sobel_new, {}))
        processing_list.append((data_item_rgb, self.document_model.get_laplace_new, {}))
        processing_list.append((data_item_rgb, self.document_model.get_gaussian_blur_new, {}))
        processing_list.append((data_item_rgb, self.document_model.get_median_filter_new, {}))
        processing_list.append((data_item_rgb, self.document_model.get_uniform_filter_new, {}))
        processing_list.append((data_item_rgb, self.document_model.get_crop_new, {}))
        processing_list.append((data_item_rgb, self.document_model.get_transpose_flip_new, {"do_transpose": True, "do_flip_v": True, "do_flip_h": True}))
        processing_list.append((data_item_rgb, self.document_model.get_resample_new, {"width": 128, "height": 128}))
        processing_list.append((data_item_rgb, self.document_model.get_histogram_new, {}))
        processing_list.append((data_item_rgb, self.document_model.get_line_profile_new, {}))
        processing_list.append((data_item_rgb, self.document_model.get_projection_new, {}))
        processing_list.append((data_item_rgb, self.document_model.get_convert_to_scalar_new, {}))

        for source_data_item, fn, params in processing_list:
            data_item = fn(source_data_item)
            for name, value in params.items():
                self.document_model.get_data_item_computation(data_item)._set_variable_value(name, value)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            self.document_model.recompute_all()
            with display_specifier.data_item.data_ref() as data_ref:
                src_data_item = self.document_model.resolve_object_specifier(self.document_model.get_data_item_computation(data_item).variables[0].variable_specifier).value.data_item
                self.assertEqual(src_data_item, source_data_item)
                self.assertIsNotNone(data_ref.data)
                self.assertIsNotNone(display_specifier.data_item.dimensional_calibrations)
                self.assertEqual(display_specifier.data_item.data_shape, data_ref.data.shape)
                self.assertEqual(display_specifier.data_item.data_dtype, data_ref.data.dtype)
                self.assertIsNotNone(display_specifier.data_item.data_dtype.type)  # make sure we're returning a dtype
                self.assertEqual(len(display_specifier.data_item.dimensional_shape), len(display_specifier.data_item.dimensional_calibrations))

    def test_processing_2d_complex128(self):
        data_item_complex = DataItem.DataItem(numpy.zeros((8, 8), numpy.complex128))
        self.document_model.append_data_item(data_item_complex)

        processing_list = []
        processing_list.append((data_item_complex, self.document_model.get_ifft_new, {}))
        processing_list.append((data_item_complex, self.document_model.get_projection_new, {}))
        processing_list.append((data_item_complex, self.document_model.get_convert_to_scalar_new, {}))

        for source_data_item, fn, params in processing_list:
            data_item = fn(source_data_item)
            for name, value in params.items():
                self.document_model.get_data_item_computation(data_item)._set_variable_value(name, value)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            self.document_model.recompute_all()
            with display_specifier.data_item.data_ref() as data_ref:
                src_data_item = self.document_model.resolve_object_specifier(self.document_model.get_data_item_computation(data_item).variables[0].variable_specifier).value.data_item
                self.assertEqual(src_data_item, source_data_item)
                self.assertIsNotNone(data_ref.data)
                self.assertIsNotNone(display_specifier.data_item.dimensional_calibrations)
                self.assertEqual(display_specifier.data_item.data_shape, data_ref.data.shape)
                self.assertEqual(display_specifier.data_item.data_dtype, data_ref.data.dtype)
                self.assertIsNotNone(display_specifier.data_item.data_dtype.type)  # make sure we're returning a dtype
                self.assertEqual(len(display_specifier.data_item.dimensional_shape), len(display_specifier.data_item.dimensional_calibrations))

    def test_processing_2d_complex64(self):
        data_item_complex = DataItem.DataItem(numpy.zeros((8, 8), numpy.complex64))
        self.document_model.append_data_item(data_item_complex)

        processing_list = []
        processing_list.append((data_item_complex, self.document_model.get_ifft_new, {}))
        processing_list.append((data_item_complex, self.document_model.get_projection_new, {}))
        processing_list.append((data_item_complex, self.document_model.get_convert_to_scalar_new, {}))

        for source_data_item, fn, params in processing_list:
            data_item = fn(source_data_item)
            for name, value in params.items():
                self.document_model.get_data_item_computation(data_item)._set_variable_value(name, value)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            self.document_model.recompute_all()
            with display_specifier.data_item.data_ref() as data_ref:
                src_data_item = self.document_model.resolve_object_specifier(self.document_model.get_data_item_computation(data_item).variables[0].variable_specifier).value.data_item
                self.assertEqual(src_data_item, source_data_item)
                self.assertIsNotNone(data_ref.data)
                self.assertIsNotNone(display_specifier.data_item.dimensional_calibrations)
                self.assertEqual(display_specifier.data_item.data_shape, data_ref.data.shape)
                self.assertEqual(display_specifier.data_item.data_dtype, data_ref.data.dtype)
                self.assertIsNotNone(display_specifier.data_item.data_dtype.type)  # make sure we're returning a dtype
                self.assertEqual(len(display_specifier.data_item.dimensional_shape), len(display_specifier.data_item.dimensional_calibrations))

    def test_processing_2d_2d_float(self):
        d = numpy.random.randn(4, 4, 3, 3)
        data_and_metadata = DataAndMetadata.new_data_and_metadata(d, data_descriptor=DataAndMetadata.DataDescriptor(False, 2, 2))
        data_item = DataItem.new_data_item(data_and_metadata)
        self.document_model.append_data_item(data_item)

        processing_list = []
        processing_list.append((data_item, self.document_model.get_projection_new, {}))
        processing_list.append((data_item, self.document_model.get_convert_to_scalar_new, {}))

        for source_data_item, fn, params in processing_list:
            data_item = fn(source_data_item)
            for name, value in params.items():
                self.document_model.get_data_item_computation(data_item)._set_variable_value(name, value)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            self.document_model.recompute_all()
            with display_specifier.data_item.data_ref() as data_ref:
                src_data_item = self.document_model.resolve_object_specifier(self.document_model.get_data_item_computation(data_item).variables[0].variable_specifier).value.data_item
                self.assertEqual(src_data_item, source_data_item)
                self.assertIsNotNone(data_ref.data)
                self.assertIsNotNone(display_specifier.data_item.dimensional_calibrations)
                self.assertEqual(display_specifier.data_item.data_shape, data_ref.data.shape)
                self.assertEqual(display_specifier.data_item.data_dtype, data_ref.data.dtype)
                self.assertIsNotNone(display_specifier.data_item.data_dtype.type)  # make sure we're returning a dtype
                self.assertEqual(len(display_specifier.data_item.dimensional_shape), len(display_specifier.data_item.dimensional_calibrations))

    # test processing against 2d data. doesn't test for correctness of the processing.
    def test_invalid_processings(self):

        data_item_none = DataItem.DataItem()
        self.document_model.append_data_item(data_item_none)

        data_item_real_0l = DataItem.DataItem(numpy.zeros((0), numpy.double))
        self.document_model.append_data_item(data_item_real_0l)
        data_item_real_0w = DataItem.DataItem(numpy.zeros((256,0), numpy.double))
        self.document_model.append_data_item(data_item_real_0w)
        data_item_real_0h = DataItem.DataItem(numpy.zeros((0,256), numpy.double))
        self.document_model.append_data_item(data_item_real_0h)
        data_item_real_0z = DataItem.DataItem(numpy.zeros((8, 8, 0), numpy.double))
        self.document_model.append_data_item(data_item_real_0z)
        data_item_real_0z0w = DataItem.DataItem(numpy.zeros((0,256,16), numpy.double))
        self.document_model.append_data_item(data_item_real_0z0w)
        data_item_real_0z0h = DataItem.DataItem(numpy.zeros((256,0,16), numpy.double))
        self.document_model.append_data_item(data_item_real_0z0h)

        data_item_complex_0l = DataItem.DataItem(numpy.zeros((0), numpy.double))
        self.document_model.append_data_item(data_item_complex_0l)
        data_item_complex_0w = DataItem.DataItem(numpy.zeros((256,0), numpy.double))
        self.document_model.append_data_item(data_item_complex_0w)
        data_item_complex_0h = DataItem.DataItem(numpy.zeros((0,256), numpy.double))
        self.document_model.append_data_item(data_item_complex_0h)
        data_item_complex_0z = DataItem.DataItem(numpy.zeros((8, 8, 0), numpy.double))
        self.document_model.append_data_item(data_item_complex_0z)
        data_item_complex_0z0w = DataItem.DataItem(numpy.zeros((0,256,16), numpy.double))
        self.document_model.append_data_item(data_item_complex_0z0w)
        data_item_complex_0z0h = DataItem.DataItem(numpy.zeros((256,0,16), numpy.double))
        self.document_model.append_data_item(data_item_complex_0z0h)

        data_item_rgb_0w = DataItem.DataItem(numpy.zeros((256,0, 3), numpy.uint8))
        self.document_model.append_data_item(data_item_rgb_0w)
        data_item_rgb_0h = DataItem.DataItem(numpy.zeros((0,256, 3), numpy.uint8))
        self.document_model.append_data_item(data_item_rgb_0h)

        data_item_rgba_0w = DataItem.DataItem(numpy.zeros((256,0, 4), numpy.uint8))
        self.document_model.append_data_item(data_item_rgba_0w)
        data_item_rgba_0h = DataItem.DataItem(numpy.zeros((0,256, 4), numpy.uint8))
        self.document_model.append_data_item(data_item_rgba_0h)

        data_list = (
            # data_item_none,
            data_item_real_0l, data_item_real_0w, data_item_real_0h, data_item_real_0z,
            data_item_real_0z0w, data_item_real_0z0h, data_item_complex_0l, data_item_complex_0w, data_item_complex_0h,
            data_item_complex_0z, data_item_complex_0z0w, data_item_complex_0z0h, data_item_rgb_0w, data_item_rgb_0h,
            data_item_rgba_0w, data_item_rgba_0h)

        processing_list = []
        for data_item in data_list:
            processing_list.append((data_item, self.document_model.get_fft_new, {}))
            processing_list.append((data_item, self.document_model.get_ifft_new, {}))
            processing_list.append((data_item, self.document_model.get_auto_correlate_new, {}))
            processing_list.append((data_item, functools.partial(self.document_model.get_cross_correlate_new, data_item), {}))
            processing_list.append((data_item, self.document_model.get_invert_new, {}))
            processing_list.append((data_item, self.document_model.get_sobel_new, {}))
            processing_list.append((data_item, self.document_model.get_laplace_new, {}))
            processing_list.append((data_item, self.document_model.get_gaussian_blur_new, {}))
            processing_list.append((data_item, self.document_model.get_median_filter_new, {}))
            processing_list.append((data_item, self.document_model.get_uniform_filter_new, {}))
            processing_list.append((data_item, self.document_model.get_crop_new, {}))
            processing_list.append((data_item, self.document_model.get_transpose_flip_new, {"do_transpose": True, "do_flip_v": True, "do_flip_h": True}))
            processing_list.append((data_item, self.document_model.get_slice_sum_new, {}))
            processing_list.append((data_item, self.document_model.get_pick_new, {}))
            processing_list.append((data_item, self.document_model.get_pick_region_new, {}))
            processing_list.append((data_item, self.document_model.get_resample_new, {"width": 128, "height": 128}))
            processing_list.append((data_item, self.document_model.get_histogram_new, {}))
            processing_list.append((data_item, self.document_model.get_line_profile_new, {}))
            processing_list.append((data_item, self.document_model.get_projection_new, {}))
            processing_list.append((data_item, self.document_model.get_convert_to_scalar_new, {}))

        for source_data_item, fn, params in processing_list:
            data_item = fn(source_data_item)
            if data_item:
                for name, value in params.items():
                    self.document_model.get_data_item_computation(data_item)._set_variable_value(name, value)
                display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
                self.document_model.recompute_all()
                with display_specifier.data_item.data_ref() as data_ref:
                    src_data_item = self.document_model.resolve_object_specifier(self.document_model.get_data_item_computation(data_item).variables[0].variable_specifier).value.data_item
                    self.assertEqual(src_data_item, source_data_item)
                    self.assertIsNone(data_ref.data)
                    self.assertFalse(display_specifier.data_item.dimensional_calibrations)


    def test_processing_on_none(self):
        # TODO: this test makes less sense with computations; but leave it here until data_item and data_item merge.
        data_item = DataItem.DataItem()
        self.document_model.append_data_item(data_item)

        processing_list = []
        processing_list.append((data_item, self.document_model.get_fft_new, {}))
        processing_list.append((data_item, self.document_model.get_ifft_new, {}))
        processing_list.append((data_item, self.document_model.get_auto_correlate_new, {}))
        processing_list.append((data_item, functools.partial(self.document_model.get_cross_correlate_new, data_item), {}))
        processing_list.append((data_item, self.document_model.get_invert_new, {}))
        processing_list.append((data_item, self.document_model.get_sobel_new, {}))
        processing_list.append((data_item, self.document_model.get_laplace_new, {}))
        processing_list.append((data_item, self.document_model.get_gaussian_blur_new, {}))
        processing_list.append((data_item, self.document_model.get_median_filter_new, {}))
        processing_list.append((data_item, self.document_model.get_uniform_filter_new, {}))
        processing_list.append((data_item, self.document_model.get_crop_new, {}))
        processing_list.append((data_item, self.document_model.get_transpose_flip_new, {"do_transpose": True, "do_flip_v": True, "do_flip_h": True}))
        processing_list.append((data_item, self.document_model.get_slice_sum_new, {}))
        processing_list.append((data_item, self.document_model.get_pick_new, {}))
        processing_list.append((data_item, self.document_model.get_pick_region_new, {}))
        processing_list.append((data_item, self.document_model.get_resample_new, {"width": 128, "height": 128}))
        processing_list.append((data_item, self.document_model.get_histogram_new, {}))
        processing_list.append((data_item, self.document_model.get_line_profile_new, {}))
        processing_list.append((data_item, self.document_model.get_projection_new, {}))
        processing_list.append((data_item, self.document_model.get_convert_to_scalar_new, {}))

        for source_data_item, fn, params in processing_list:
            data_item = fn(source_data_item)
            if data_item:
                for name, value in params.items():
                    self.document_model.get_data_item_computation(data_item)._set_variable_value(name, value)
                display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
                self.document_model.recompute_all()
                with display_specifier.data_item.data_ref() as data_ref:
                    src_data_item = self.document_model.resolve_object_specifier(self.document_model.get_data_item_computation(data_item).variables[0].variable_specifier).value.data_item
                    self.assertEqual(src_data_item, source_data_item)
                    self.assertIsNone(data_ref.data)
                    self.assertEqual(display_specifier.data_item.dimensional_calibrations, [])

    def test_crop_2d_processing_returns_correct_dimensional_shape_and_data_shape(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item = DataItem.DataItem(numpy.zeros((20,10), numpy.double))
            crop_region = Graphics.RectangleGraphic()
            crop_region.bounds = (0.2, 0.3), (0.5, 0.5)
            data_item.displays[0].add_graphic(crop_region)
            document_model.append_data_item(data_item)
            real_data_item = document_model.get_crop_new(data_item, crop_region)
            real_display_specifier = DataItem.DisplaySpecifier.from_data_item(real_data_item)
            document_model.recompute_all()
            # make sure we get the right shape
            self.assertEqual(real_display_specifier.data_item.dimensional_shape, (10, 5))
            with real_display_specifier.data_item.data_ref() as data_real_accessor:
                self.assertEqual(data_real_accessor.data.shape, (10, 5))

    def test_fft_2d_dtype(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item = DataItem.DataItem(numpy.zeros((16, 16), numpy.float64))
            document_model.append_data_item(data_item)
            fft_data_item = document_model.get_fft_new(data_item)
            fft_display_specifier = DataItem.DisplaySpecifier.from_data_item(fft_data_item)
            document_model.recompute_all()
            with fft_display_specifier.data_item.data_ref() as fft_data_ref:
                self.assertEqual(fft_data_ref.data.shape, (16, 16))
                self.assertEqual(fft_data_ref.data.dtype, numpy.dtype(numpy.complex128))

    def test_convert_complex128_to_scalar_results_in_float64(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item = DataItem.DataItem(numpy.zeros((16, 16), numpy.complex128))
            document_model.append_data_item(data_item)
            scalar_data_item = document_model.get_convert_to_scalar_new(data_item)
            scalar_display_specifier = DataItem.DisplaySpecifier.from_data_item(scalar_data_item)
            document_model.recompute_all()
            with scalar_display_specifier.data_item.data_ref() as scalar_data_ref:
                self.assertEqual(scalar_data_ref.data.dtype, numpy.dtype(numpy.float64))

    def test_rgba_invert_processing_should_retain_alpha(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            rgba_data_item = DataItem.DataItem(numpy.zeros((8, 8, 4), numpy.uint8))
            document_model.append_data_item(rgba_data_item)
            rgba_display_specifier = DataItem.DisplaySpecifier.from_data_item(rgba_data_item)
            with rgba_display_specifier.data_item.data_ref() as data_ref:
                data_ref.master_data[:] = (20,40,60,100)
                data_ref.master_data_updated()
            rgba2_data_item = document_model.get_invert_new(rgba_data_item)
            rgba2_display_specifier = DataItem.DisplaySpecifier.from_data_item(rgba2_data_item)
            document_model.recompute_all()
            with rgba2_display_specifier.data_item.data_ref() as data_ref:
                pixel = data_ref.data[0,0,...]
                self.assertEqual(pixel[0], 255 - 20)
                self.assertEqual(pixel[1], 255 - 40)
                self.assertEqual(pixel[2], 255 - 60)
                self.assertEqual(pixel[3], 100)

    def test_deepcopy_of_crop_processing_should_copy_roi(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item_rgba = DataItem.DataItem(numpy.zeros((8, 8, 4), numpy.uint8))
            crop_region = Graphics.RectangleGraphic()
            crop_region.bounds = (0.25, 0.25), (0.5, 0.5)
            data_item_rgba.displays[0].add_graphic(crop_region)
            self.document_model.append_data_item(data_item_rgba)
            data_item_rgba2 = document_model.get_crop_new(data_item_rgba, crop_region)
            data_item_rgba2_copy = copy.deepcopy(data_item_rgba2)
            # make sure the computation was not copied
            self.assertNotEqual(document_model.get_data_item_computation(data_item_rgba2), document_model.get_data_item_computation(data_item_rgba2_copy))

    def test_snapshot_of_processing_should_copy_data(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item_rgba = DataItem.DataItem(numpy.zeros((8, 8, 4), numpy.uint8))
            document_model.append_data_item(data_item_rgba)
            data_item_rgba2 = document_model.get_invert_new(data_item_rgba)
            document_model.recompute_all()
            data_item_rgba2_ss = document_model.get_snapshot_new(data_item_rgba2)
            self.assertTrue(data_item_rgba2_ss.has_data)

    def test_snapshot_empty_data_item_should_produce_empty_data_item(self):
        data_item = DataItem.DataItem()
        data_item.ensure_data_source()
        display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
        self.assertIsNone(display_specifier.data_item.data)
        self.assertIsNone(display_specifier.data_item.data_dtype)
        self.assertIsNone(display_specifier.data_item.data_shape)
        snapshot_data_item = data_item.snapshot()
        snapshot_display_specifier = DataItem.DisplaySpecifier.from_data_item(snapshot_data_item)
        self.assertIsNone(snapshot_display_specifier.data_item.data)
        self.assertIsNone(snapshot_display_specifier.data_item.data_dtype)
        self.assertIsNone(snapshot_display_specifier.data_item.data_shape)

    def test_snapshot_of_processing_should_copy_calibrations_not_dimensional_calibrations(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            # setup
            display_specifier.data_item.set_dimensional_calibration(0, Calibration.Calibration(5.0, 2.0, u"nm"))
            display_specifier.data_item.set_dimensional_calibration(1, Calibration.Calibration(5.0, 2.0, u"nm"))
            display_specifier.data_item.set_intensity_calibration(Calibration.Calibration(7.5, 2.5, u"ll"))
            data_item2 = document_model.get_invert_new(data_item)
            display_specifier2 = DataItem.DisplaySpecifier.from_data_item(data_item2)
            document_model.recompute_all()
            # make sure our assumptions are correct
            self.assertEqual(len(display_specifier.data_item.dimensional_calibrations), 2)
            self.assertEqual(len(display_specifier2.data_item.dimensional_calibrations), 2)
            # take snapshot
            snapshot_data_item = document_model.get_snapshot_new(data_item2)
            # check calibrations
            self.assertEqual(len(snapshot_data_item.dimensional_calibrations), 2)
            self.assertEqual(snapshot_data_item.dimensional_calibrations[0].scale, 2.0)
            self.assertEqual(snapshot_data_item.dimensional_calibrations[0].offset, 5.0)
            self.assertEqual(snapshot_data_item.dimensional_calibrations[0].units, u"nm")
            self.assertEqual(snapshot_data_item.dimensional_calibrations[1].scale, 2.0)
            self.assertEqual(snapshot_data_item.dimensional_calibrations[1].offset, 5.0)
            self.assertEqual(snapshot_data_item.dimensional_calibrations[1].units, u"nm")
            self.assertEqual(snapshot_data_item.intensity_calibration.scale, 2.5)
            self.assertEqual(snapshot_data_item.intensity_calibration.offset, 7.5)
            self.assertEqual(snapshot_data_item.intensity_calibration.units, u"ll")

    def test_crop_2d_processing_on_calibrated_data_results_in_calibration_with_correct_offset(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item = DataItem.DataItem(numpy.zeros((20, 10), numpy.double))
            crop_region = Graphics.RectangleGraphic()
            crop_region.bounds = (0.2, 0.3), (0.5, 0.5)
            data_item.displays[0].add_graphic(crop_region)
            document_model.append_data_item(data_item)
            spatial_calibration_0 = data_item.dimensional_calibrations[0]
            spatial_calibration_0.offset = 20.0
            spatial_calibration_0.scale = 5.0
            spatial_calibration_0.units = "dogs"
            spatial_calibration_1 = data_item.dimensional_calibrations[1]
            spatial_calibration_1.offset = 55.0
            spatial_calibration_1.scale = 5.5
            spatial_calibration_1.units = "cats"
            data_item.set_dimensional_calibration(0, spatial_calibration_0)
            data_item.set_dimensional_calibration(1, spatial_calibration_1)
            data_item2 = document_model.get_crop_new(data_item, crop_region)
            display_specifier2 = DataItem.DisplaySpecifier.from_data_item(data_item2)
            document_model.recompute_all()
            # make sure the calibrations are correct
            self.assertAlmostEqual(display_specifier2.data_item.dimensional_calibrations[0].offset, 20.0 + 20 * 0.2 * 5.0)
            self.assertAlmostEqual(display_specifier2.data_item.dimensional_calibrations[1].offset, 55.0 + 10 * 0.3 * 5.5)
            self.assertAlmostEqual(display_specifier2.data_item.dimensional_calibrations[0].scale, 5.0)
            self.assertAlmostEqual(display_specifier2.data_item.dimensional_calibrations[1].scale, 5.5)
            self.assertEqual(display_specifier2.data_item.dimensional_calibrations[0].units, "dogs")
            self.assertEqual(display_specifier2.data_item.dimensional_calibrations[1].units, "cats")

    def test_projection_2d_processing_on_calibrated_data_results_in_calibration_with_correct_offset(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item = DataItem.DataItem(numpy.zeros((20, 10), numpy.double))
            document_model.append_data_item(data_item)
            spatial_calibration_0 = data_item.dimensional_calibrations[0]
            spatial_calibration_0.offset = 20.0
            spatial_calibration_0.scale = 5.0
            spatial_calibration_0.units = "dogs"
            spatial_calibration_1 = data_item.dimensional_calibrations[1]
            spatial_calibration_1.offset = 55.0
            spatial_calibration_1.scale = 5.5
            spatial_calibration_1.units = "cats"
            data_item.set_dimensional_calibration(0, spatial_calibration_0)
            data_item.set_dimensional_calibration(1, spatial_calibration_1)
            data_item2 = document_model.get_projection_new(data_item)
            display_specifier2 = DataItem.DisplaySpecifier.from_data_item(data_item2)
            document_model.recompute_all()
            # make sure the calibrations are correct
            self.assertAlmostEqual(display_specifier2.data_item.dimensional_calibrations[0].offset, 55.0)
            self.assertAlmostEqual(display_specifier2.data_item.dimensional_calibrations[0].scale, 5.5)
            self.assertEqual(display_specifier2.data_item.dimensional_calibrations[0].units, "cats")

    def disabled_test_removing_computation_with_multiple_associated_regions_removes_all_regions(self):
        self.assertFalse(True)

    def test_crop_works_on_selected_region_without_exception(self):
        document_model = DocumentModel.DocumentModel()
        data_item = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
        document_model.append_data_item(data_item)
        display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        display_panel = document_controller.selected_display_panel
        display_panel.set_display_panel_data_item(data_item)
        self.assertEqual(len(display_specifier.display.graphics), 0)
        crop_region = Graphics.RectangleGraphic()
        crop_region.center = (0.5, 0.5)
        crop_region.size = (0.5, 1.0)
        data_item.displays[0].add_graphic(crop_region)
        display_specifier.display.graphic_selection.set(0)
        document_controller.processing_crop().data_item
        document_controller.close()

    def test_remove_graphic_for_crop_removes_processed_data_item(self):
        document_model = DocumentModel.DocumentModel()
        data_item = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
        document_model.append_data_item(data_item)
        display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        display_panel = document_controller.selected_display_panel
        display_panel.set_display_panel_data_item(data_item)
        self.assertEqual(len(display_specifier.display.graphics), 0)
        crop_region = Graphics.RectangleGraphic()
        crop_region.center = (0.5, 0.5)
        crop_region.size = (0.5, 1.0)
        data_item.displays[0].add_graphic(crop_region)
        display_specifier.display.graphic_selection.set(0)
        cropped_data_item = document_controller.processing_crop().data_item
        document_controller.periodic()  # TODO: remove need to let the inspector catch up
        self.assertEqual(len(display_specifier.display.graphics), 1)
        self.assertTrue(cropped_data_item in document_model.data_items)
        display_specifier.display.graphic_selection.clear()
        display_specifier.display.graphic_selection.add(0)
        # make sure assumptions are correct
        self.assertEqual(document_model.get_source_data_items(cropped_data_item)[0], data_item)
        self.assertTrue(cropped_data_item in document_model.data_items)
        # remove the graphic and make sure things are as expected
        display_panel.set_display_panel_data_item(data_item)
        document_controller.remove_selected_graphics()
        self.assertEqual(len(display_specifier.display.graphics), 0)
        self.assertEqual(len(display_specifier.display.graphic_selection.indexes), 0)  # disabled until test_remove_line_profile_updates_graphic_selection
        self.assertFalse(cropped_data_item in document_model.data_items)
        # clean up
        document_controller.close()

    def test_remove_graphic_for_crop_combined_with_another_processing_removes_processed_data_item(self):
        document_model = DocumentModel.DocumentModel()
        data_item = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
        document_model.append_data_item(data_item)
        display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        display_panel = document_controller.selected_display_panel
        display_panel.set_display_panel_data_item(data_item)
        self.assertEqual(len(display_specifier.display.graphics), 0)
        crop_region = Graphics.RectangleGraphic()
        crop_region.center = (0.5, 0.5)
        crop_region.size = (0.5, 1.0)
        data_item.displays[0].add_graphic(crop_region)
        display_specifier.display.graphic_selection.set(0)
        projection_data_item = document_controller.processing_projection().data_item
        document_controller.periodic()
        self.assertTrue(projection_data_item in document_model.data_items)
        display_specifier.display.graphic_selection.clear()
        display_specifier.display.graphic_selection.add(0)
        # make sure assumptions are correct
        self.assertEqual(document_model.get_source_data_items(projection_data_item)[0], data_item)
        self.assertTrue(projection_data_item in document_model.data_items)
        # remove the graphic and make sure things are as expected
        display_panel.set_display_panel_data_item(data_item)
        document_controller.remove_selected_graphics()
        self.assertEqual(len(display_specifier.display.graphics), 0)
        self.assertEqual(len(display_specifier.display.graphic_selection.indexes), 0)  # disabled until test_remove_line_profile_updates_graphic_selection
        self.assertFalse(projection_data_item in document_model.data_items)
        # clean up
        document_controller.close()

    def test_modifying_processing_results_in_data_computation(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            # set up the data items
            data_item = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            document_model.append_data_item(data_item)
            blurred_data_item = document_model.get_gaussian_blur_new(data_item)
            blurred_display_specifier = DataItem.DisplaySpecifier.from_data_item(blurred_data_item)
            # establish listeners
            data_changed_ref = [False]
            display_changed_ref = [False]
            def data_item_content_changed():
                data_changed_ref[0] = True
            with contextlib.closing(blurred_data_item.library_item_changed_event.listen(data_item_content_changed)):
                def display_changed():
                    display_changed_ref[0] = True
                with contextlib.closing(blurred_display_specifier.display.display_changed_event.listen(display_changed)):
                    # modify processing. make sure data and dependent data gets updated.
                    data_changed_ref[0] = False
                    display_changed_ref[0] = False
                    document_model.get_data_item_computation(blurred_data_item)._set_variable_value("sigma", 0.1)
                    document_model.recompute_all()
                    self.assertTrue(data_changed_ref[0])
                    self.assertTrue(display_changed_ref[0])

    def test_modifying_processing_region_results_in_data_computation(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            # set up the data items
            data_item = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            crop_region = Graphics.RectangleGraphic()
            crop_region.bounds = (0.25, 0.25), (0.5, 0.5)
            data_item.displays[0].add_graphic(crop_region)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            document_model.append_data_item(data_item)
            cropped_data_item = document_model.get_crop_new(data_item, crop_region)
            cropped_display_specifier = DataItem.DisplaySpecifier.from_data_item(cropped_data_item)
            # establish listeners
            data_changed_ref = [False]
            display_changed_ref = [False]
            def data_item_content_changed():
                data_changed_ref[0] = True
            with contextlib.closing(cropped_data_item.library_item_changed_event.listen(data_item_content_changed)):
                def display_changed():
                    display_changed_ref[0] = True
                with contextlib.closing(cropped_display_specifier.display.display_changed_event.listen(display_changed)):
                    # modify processing. make sure data and dependent data gets updated.
                    data_changed_ref[0] = False
                    display_changed_ref[0] = False
                    crop_region.center = (0.51, 0.51)
                    document_model.recompute_all()
                    self.assertTrue(data_changed_ref[0])
                    self.assertTrue(display_changed_ref[0])

    def test_changing_region_does_not_trigger_fft_recompute(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            crop_region = Graphics.RectangleGraphic()
            crop_region.bounds = (0.25, 0.25), (0.5, 0.5)
            data_item.displays[0].add_graphic(crop_region)
            document_model.append_data_item(data_item)
            fft_data_item = document_model.get_fft_new(data_item)
            crop_data_item = document_model.get_crop_new(data_item, crop_region)
            document_model.recompute_all()
            self.assertFalse(document_model.get_data_item_computation(fft_data_item).needs_update)
            self.assertFalse(document_model.get_data_item_computation(crop_data_item).needs_update)
            crop_region.bounds = Geometry.FloatRect(crop_region.bounds[0], Geometry.FloatPoint(0.1, 0.1))
            self.assertTrue(document_model.get_data_item_computation(crop_data_item).needs_update)
            self.assertFalse(document_model.get_data_item_computation(fft_data_item).needs_update)

    def test_removing_source_of_cross_correlation_does_not_throw_exception(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item1 = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            data_item2 = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            document_model.append_data_item(data_item1)
            document_model.append_data_item(data_item2)
            document_model.get_cross_correlate_new(data_item1, data_item2)
            document_model.recompute_all()
            document_model.remove_data_item(data_item1)
            document_model.recompute_all()

    def test_crop_of_slice_of_3d_handles_dimensions(self):
        # the bug was that slice processing returned the wrong number of dimensions
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item = DataItem.DataItem(numpy.zeros((32, 32, 16), numpy.float))
            document_model.append_data_item(data_item)
            slice_data_item = document_model.get_slice_sum_new(data_item)
            document_model.recompute_all()
            crop_region = Graphics.RectangleGraphic()
            crop_region.bounds = (0.25, 0.25), (0.5, 0.5)
            slice_data_item.displays[0].add_graphic(crop_region)
            crop_data_item = document_model.get_crop_new(slice_data_item, crop_region)
            document_model.recompute_all()

    def test_projection_of_2d_by_2d_sums_first_datum_dimension(self):
        # the bug was that slice processing returned the wrong number of dimensions
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            d = numpy.random.randn(4, 4, 3, 3)
            data_and_metadata = DataAndMetadata.new_data_and_metadata(d, data_descriptor=DataAndMetadata.DataDescriptor(False, 2, 2))
            data_item = DataItem.new_data_item(data_and_metadata)
            document_model.append_data_item(data_item)
            slice_data_item = document_model.get_projection_new(data_item)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(numpy.sum(d, 2), slice_data_item.xdata.data))

    def test_cross_correlate_works_in_1d(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data = ((numpy.abs(numpy.random.randn(8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(data)
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(data_item)
            document_controller.processing_cross_correlate_new()

    def test_cross_correlate_works_in_2d(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data = ((numpy.abs(numpy.random.randn(8, 8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(data)
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(data_item)
            document_controller.processing_cross_correlate_new()

    def test_get_two_data_sources_handles_no_selection(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            self.assertIsNone(document_controller._get_two_data_sources())
            data = ((numpy.abs(numpy.random.randn(8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            self.assertIsNone(document_controller._get_two_data_sources())

    def test_get_two_data_sources_handles_one_selected_data_item(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            self.assertIsNone(document_controller._get_two_data_sources())
            data = ((numpy.abs(numpy.random.randn(8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            self.assertIsNone(document_controller._get_two_data_sources())
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(data_item)
            data_sources = document_controller._get_two_data_sources()
            self.assertEqual(data_sources[0][0], data_item)
            self.assertEqual(data_sources[0][1], None)
            self.assertEqual(data_sources[1][0], data_item)
            self.assertEqual(data_sources[1][1], None)

    def test_get_two_data_sources_handles_two_selected_data_items(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            self.assertIsNone(document_controller._get_two_data_sources())
            data = ((numpy.abs(numpy.random.randn(8)) + 1) * 10).astype(numpy.uint32)
            data_item1 = DataItem.DataItem(data)
            document_model.append_data_item(data_item1)
            data_item2 = DataItem.DataItem(numpy.random.randn(8))
            document_model.append_data_item(data_item2)
            self.assertIsNone(document_controller._get_two_data_sources())
            document_controller.select_data_items_in_data_panel([data_item1, data_item2])
            data_sources = document_controller._get_two_data_sources()
            self.assertEqual(data_sources[0][0], data_item1)
            self.assertEqual(data_sources[0][1], None)
            self.assertEqual(data_sources[1][0], data_item2)
            self.assertEqual(data_sources[1][1], None)

    def test_get_two_data_sources_handles_three_selected_data_items(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_items = list()
            for i in range(3):
                data_item = DataItem.DataItem(numpy.random.randn(8))
                document_model.append_data_item(data_item)
            document_controller.select_data_items_in_data_panel(document_model.data_items)
            self.assertIsNone(document_controller._get_two_data_sources())

    def test_get_two_data_sources_handles_one_selected_data_item_and_one_crop_graphic(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            self.assertIsNone(document_controller._get_two_data_sources())
            data = ((numpy.abs(numpy.random.randn(8, 8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(data)
            display = data_item.displays[0]
            crop_region = Graphics.RectangleGraphic()
            crop_region.center = (0.5, 0.5)
            crop_region.size = (0.5, 1.0)
            display.add_graphic(crop_region)
            document_model.append_data_item(data_item)
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(data_item)
            display.graphic_selection.set(0)
            data_sources = document_controller._get_two_data_sources()
            self.assertEqual(data_sources[0][0], data_item)
            self.assertEqual(data_sources[0][1], crop_region)
            self.assertEqual(data_sources[1][0], data_item)
            self.assertEqual(data_sources[1][1], crop_region)

    def test_get_two_data_sources_handles_one_selected_data_item_and_two_crop_graphics(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            self.assertIsNone(document_controller._get_two_data_sources())
            data = ((numpy.abs(numpy.random.randn(8, 8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(data)
            display = data_item.displays[0]
            crop_region1 = Graphics.RectangleGraphic()
            crop_region1.center = (0.5, 0.5)
            crop_region1.size = (0.5, 1.0)
            display.add_graphic(crop_region1)
            crop_region2 = Graphics.RectangleGraphic()
            crop_region2.center = (0.6, 0.5)
            crop_region2.size = (0.5, 1.0)
            display.add_graphic(crop_region2)
            document_model.append_data_item(data_item)
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(data_item)
            display.graphic_selection.set(0)
            display.graphic_selection.add(1)
            data_sources = document_controller._get_two_data_sources()
            self.assertEqual(data_sources[0][0], data_item)
            self.assertEqual(data_sources[0][1], crop_region1)
            self.assertEqual(data_sources[1][0], data_item)
            self.assertEqual(data_sources[1][1], crop_region2)

    def test_get_two_data_sources_handles_one_selected_data_item_and_three_crop_graphics(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            self.assertIsNone(document_controller._get_two_data_sources())
            data = ((numpy.abs(numpy.random.randn(8, 8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(data)
            display = data_item.displays[0]
            for i in range(3):
                display.add_graphic(Graphics.RectangleGraphic())
            document_model.append_data_item(data_item)
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(data_item)
            display.graphic_selection.add_range(range(3))
            data_sources = document_controller._get_two_data_sources()
            self.assertEqual(data_sources[0][0], data_item)
            self.assertEqual(data_sources[0][1], None)
            self.assertEqual(data_sources[1][0], data_item)
            self.assertEqual(data_sources[1][1], None)

    def test_crop_handles_1d_case_with_existing_interval(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data = ((numpy.abs(numpy.random.randn(8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            crop_region = Graphics.IntervalGraphic()
            crop_region.interval = (0.5, 1.0)
            display = data_item.displays[0]
            display.add_graphic(crop_region)
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(data_item)
            display.graphic_selection.set(0)
            cropped_data_item = document_controller.processing_crop().data_item
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(cropped_data_item.data, data_item.data[4:8]))

    def test_crop_handles_2d_case_with_existing_rectangle(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data = ((numpy.abs(numpy.random.randn(8, 8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            crop_region = Graphics.RectangleGraphic()
            crop_region.center = (0.5, 0.5)
            crop_region.size = (0.5, 1.0)
            display = data_item.displays[0]
            display.add_graphic(crop_region)
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(data_item)
            display.graphic_selection.set(0)
            cropped_data_item = document_controller.processing_crop().data_item
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(cropped_data_item.data, data_item.data[2:6, 0:8]))


if __name__ == '__main__':
    logging.getLogger().setLevel(logging.DEBUG)
    unittest.main()



import json
import unittest

from nion.swift.model import Utility

class TestUtilityClass(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_backwards_compatibility_for_short_versions(self):
        self.assertEqual(Utility.compare_versions("1", "1.0.0"), 0)
        self.assertEqual(Utility.compare_versions("1", "1.0.4"), 0)
        self.assertLess(Utility.compare_versions("1", "1.1.0"), 0)

    def test_compatible_version(self):
        self.assertEqual(Utility.compare_versions("~1.0", "1.0.0"), 0)
        self.assertGreater(Utility.compare_versions("~1.1", "1.0.0"), 0)
        self.assertEqual(Utility.compare_versions("~1.1", "1.1.0"), 0)
        self.assertEqual(Utility.compare_versions("~1.1", "1.1.4"), 0)
        self.assertLess(Utility.compare_versions("~1.1", "1.2.0"), 0)

    def test_incompatible_version(self):
        self.assertLess(Utility.compare_versions("~1.0", "2.0.0"), 0)
        self.assertGreater(Utility.compare_versions("~2.0", "1.0.0"), 0)

    def test_invalid_compatible_version(self):
        with self.assertRaises(Exception):
            Utility.compare_versions("~1", "1.0.0")

    def test_clean_dict_handles_none_in_tuples_and_lists(self):
        d0 = {"abc": (None, 2)}
        d1 = {"abc": (2, None)}
        d2 = {"abc": [None, 2]}
        d3 = {"abc": [2, None]}
        self.assertEqual(Utility.clean_dict(d0), d0)
        self.assertEqual(Utility.clean_dict(d1), d1)
        self.assertEqual(Utility.clean_dict(d2), d2)
        self.assertEqual(Utility.clean_dict(d3), d3)
        # ok for json to switch tuples to lists
        self.assertEqual(Utility.clean_dict(json.loads(json.dumps(d0))), d2)
        self.assertEqual(Utility.clean_dict(json.loads(json.dumps(d1))), d3)


if __name__ == '__main__':
    unittest.main()



# futures
from __future__ import absolute_import

# standard libraries
import os
import unittest

# local libraries
from nion.swift import Decorators


class TestDecoratorsClass(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_relative_file(self):
        f = Decorators.relative_file(__file__, "test.dat")
        self.assertEqual(os.path.basename(f), "test.dat")
        self.assertEqual(os.path.dirname(os.path.abspath(__file__)), os.path.dirname(f))